@@ -1,7 +1,7 @@
 package Pod::Weaver::Plugin::Bencher::Scenario;
 
 our $DATE = '2016-01-05'; # DATE
-our $VERSION = '0.02'; # VERSION
+our $VERSION = '0.03'; # VERSION
 
 use 5.010001;
 use Moose;
@@ -11,6 +11,15 @@ with 'Pod::Weaver::Role::Section';
 use Perinci::Result::Format::Lite;
 use Bencher;
 
+sub __ver_or_vers {
+    my $v = shift;
+    if (ref($v) eq 'ARRAY') {
+        return join(", ", @$v);
+    } else {
+        return $v;
+    }
+}
+
 sub _md2pod {
     require Markdown::To::POD;
 
@@ -90,52 +99,37 @@ sub _process_module {
 
     my @modules = Bencher::_get_participant_modules($scenario);
 
-    # add Benchmarked Modules section
-    {
-        last unless @modules;
-        my $pod = join('', map {"L<$_>\n\n"} @modules);
-        $self->add_text_to_section(
-            $document, $pod, 'BENCHMARKED MODULES',
-            {
-                after_section => 'SYNOPSIS',
-                before_section => 'DESCRIPTION',
-            });
-        # XXX if each participant is a unique module, then list with BENCHMARKED
-        # MODULES as above. if there is a module which has two+ participants,
-        # list like: *) L<Foo::Bar>'s C<routine1()>; *) C<Foo::Bar>'s
-        # C<routine2()>.
-    }
-
     # add Sample Benchmark Results section
+    my ($bench_res, $fres, $f2res);
     {
         my @pod;
-        my ($res, $fres, $f2res);
 
-        $res = Bencher::bencher(
+        $bench_res = Bencher::bencher(
             action => 'bench',
             scenario_module => $scenario_name,
         );
-        $fres = Bencher::format_result($res);
-        $f2res = Perinci::Result::Format::Lite::format($res, 'text-pretty');
+        $fres = Bencher::format_result($bench_res);
+        $f2res = Perinci::Result::Format::Lite::format($fres, 'text-pretty');
         $f2res =~ s/^/ /gm;
 
-        my $num_cores = $res->[3]{'func.cpu_info'}[0]{number_of_cores};
+        my $num_cores = $bench_res->[3]{'func.cpu_info'}[0]{number_of_cores};
         push @pod, "Run on: ",
-            "CPU: I<< ", $res->[3]{'func.cpu_info'}[0]{name}, " ($num_cores cores) >>, ",
-            "OS: I<< ", $res->[3]{'func.platform_info'}{osname}, " ", $res->[3]{'func.platform_info'}{oslabel}, " version ", $res->[3]{'func.platform_info'}{osvers}, " >>, ",
-            "OS kernel: I<< ", $res->[3]{'func.platform_info'}{kname}, " version ", $res->[3]{'func.platform_info'}{kvers}, " >>",
+            "perl: I<< ", __ver_or_vers($bench_res->[3]{'func.module_versions'}{perl}), " >>, ",
+            "CPU: I<< ", $bench_res->[3]{'func.cpu_info'}[0]{name}, " ($num_cores cores) >>, ",
+            "OS: I<< ", $bench_res->[3]{'func.platform_info'}{osname}, " ", $bench_res->[3]{'func.platform_info'}{oslabel}, " version ", $bench_res->[3]{'func.platform_info'}{osvers}, " >>, ",
+            "OS kernel: I<< ", $bench_res->[3]{'func.platform_info'}{kname}, " version ", $bench_res->[3]{'func.platform_info'}{kvers}, " >>",
             ".\n\n";
 
         push @pod, "Benchmark with default option:\n\n", $f2res, "\n\n";
 
         if (@modules && !$scenario->{module_startup}) {
-            $res = Bencher::bencher(
+            $bench_res = Bencher::bencher(
                 action => 'bench',
                 module_startup => 1,
                 scenario_module => $scenario_name,
             );
-            $fres = Bencher::format_result($res);
-            $f2res = Perinci::Result::Format::Lite::format($res, 'text-pretty');
+            $fres = Bencher::format_result($bench_res);
+            $f2res = Perinci::Result::Format::Lite::format($fres, 'text-pretty');
             $f2res =~ s/^/ /gm;
             push @pod, "Benchmark module startup overhead:\n\n", $f2res, "\n\n";
         }
@@ -148,6 +142,74 @@ sub _process_module {
             });
     }
 
+    # add Benchmarked Modules section
+    {
+        last unless @modules;
+        my @pod;
+
+        for my $mod (@modules) {
+            push @pod, "L<$mod>";
+            my $v = $bench_res->[3]{'func.module_versions'}{$mod};
+            if ($v) {
+                push @pod, " ", __ver_or_vers($v);
+            }
+            push @pod, "\n\n";
+        }
+
+        $self->add_text_to_section(
+            $document, join("", @pod), 'BENCHMARKED MODULES',
+            {
+                after_section => 'SYNOPSIS',
+                before_section => ['SAMPLE BENCHMARK RESULTS', 'DESCRIPTION'],
+            });
+    }
+
+    # add Benchmark Participants section
+    {
+        my @pod;
+        my $res = Bencher::bencher(
+            action => 'list-participants',
+            scenario_module => $scenario_name,
+            detail => 1,
+        );
+        push @pod, "=over\n\n";
+        my $i = -1;
+        for my $p (@{ $res->[2] }) {
+            $i++;
+            my $p0 = $scenario->{participants}[$i];
+            push @pod, "=item * ", ($p->{name} // ''), " ($p->{type})",
+                ($p->{include_by_default} ? "" : " (not included by default)"),
+                "\n\n";
+            if ($p0->{summary}) {
+                push @pod, $p0->{summary}, ".\n\n";
+            }
+            if ($p->{cmdline}) {
+                push @pod, "Command line:\n\n", " $p->{cmdline}\n\n";
+            } elsif ($p0->{fcall_template}) {
+                my $val = $p0->{fcall_template}; $val =~ s/^/ /gm;
+                push @pod, "Function call template:\n\n", $val, "\n\n";
+            } elsif ($p0->{code_template}) {
+                my $val = $p0->{code_template}; $val =~ s/^/ /gm;
+                push @pod, "Code template:\n\n", $val, "\n\n";
+            } elsif ($p->{module}) {
+                push @pod, "L<$p->{module}>";
+                if ($p->{function}) {
+                    push @pod, "::$p->{function}";
+                }
+                push @pod, "\n\n";
+            }
+            push @pod, "\n\n";
+        }
+        push @pod, "=back\n\n";
+
+        $self->add_text_to_section(
+            $document, join("", @pod), 'BENCHMARK PARTICIPANTS',
+            {
+                after_section => ['BENCHMARKED MODULES'],
+                before_section => ['SAMPLE BENCHMARK RESULTS'],
+            });
+    }
+
     $self->log(["Generated POD for '%s'", $filename]);
 }
 
@@ -179,7 +241,7 @@ Pod::Weaver::Plugin::Bencher::Scenario - Put various information from scenario i
 
 =head1 VERSION
 
-This document describes version 0.02 of Pod::Weaver::Plugin::Bencher::Scenario (from Perl distribution Pod-Weaver-Plugin-Bencher-Scenario), released on 2016-01-05.
+This document describes version 0.03 of Pod::Weaver::Plugin::Bencher::Scenario (from Perl distribution Pod-Weaver-Plugin-Bencher-Scenario), released on 2016-01-05.
 
 =head1 SYNOPSIS
 
